"""
This module contains the `Damj` class, which is the main class for the `damj` package.
"""
import os
from typing import List, Dict, Union
from IPython.display import Markdown
import pyperclip
from damj.utils import get_project_structure, matches_pattern, get_file_content, show_markdown

class Damj:
    """
    Damj

    This class provides the main functionality for the `damj` package. The `Damj` class is
    designed to help developers create prompts for large language models (LLMs) with ease
    by combining different project files within the developer's working directory.

    The package supports various customization options for including or excluding certain
    types of content from Python files and Jupyter notebooks.

    You can read the Damj documentation on GitHub [HERE](https://github.com/baselhusam/damj).

    You can use the `Damj` class with the following methods:
    ```python
    import os
    from damj import Damj

    cwd = os.getcwd()
    damj = Damj(cwd)
    damj.project_info(
        project_overview="This is a sample project.",
        add_project_structure=True,
    )
    prompt = damj.create_prompt(
        question="What is the purpose of this project?",
    )
    print(prompt)
    ```

    ---

    Also, you can have the following customization options:

    ```python
    import os
    from damj import Damj

    cwd = os.getcwd()
    damj = Damj(
    cwd=cwd,
    whitelist_files=["*.py"],
    blacklist_files=[".venv", "__pycache__"],
    snippet_marker="```"
    )

    damj.project_info(
        project_overview="This is a sample project.",
        add_project_structure=True,
        add_files_content=True,
        py_options={
            "add_imports": True,
            "add_comments": True,
            "add_docstrings": False,
            "ipynb_output": False
        }
    )

    prompt = damj.create_prompt(
        question="What is the purpose of this project?",
        copy_to_clipboard=True,
        to_markdown=False
    )

    print(prompt)
    ```

    You can customize the `Damj` class by specifying the following parameters:
    - cwd: The current working directory
    - whitelist_files: A list of files to include in the prompt
    - blacklist_files: A list of files to exclude from the prompt
    - snippet_marker: The marker used to separate code snippets in the prompt

    Parameters:
    ----------
    cwd : str
        The current working directory
    whitelist_files : List[str], optional
        A list of files to include in the prompt, by default None
    blacklist_files : List[str], optional
        A list of files to exclude from the prompt, by default None
    snippet_marker : str, optional
        The marker used to separate code snippets in the prompt, by default "```"

    Attributes:
    ----------
    cwd : str
        The current working directory
    whitelist_files : List[str]
        A list of files to include in the prompt
    blacklist_files : List[str]
        A list of files to exclude from the prompt
    snippet_marker : str
        The marker used to separate code snippets in the prompt
    prompt : str
        The prompt generated by the `Damj` class

    Methods:
    --------

    ```python
    project_info(
        project_overview: str = "",
        add_project_structure: bool = False,
        add_files_content: bool = False,
        py_options: Union[Dict, None] = None,
    ) -> str:
    ```
        Generate the project information for the prompt

    ```python
    create_prompt(
        question: Union[str, None] = None,
        copy_to_clipboard: bool = True,
        to_markdown: bool = False,
    ) -> Union[str, Markdown]:
    ```
        Create the prompt for the large language model (LLM)
    """
    def __init__(
        self,
        cwd: str,
        whitelist_files: List[str] = None,
        blacklist_files: List[str] = None,
        snippet_marker: str = "```",
    ):
        self.cwd = cwd
        whitelist_files = ["*"] if whitelist_files is None else whitelist_files
        self.blacklist_files = ["__pycache__"] if blacklist_files is None else blacklist_files
        self.whitelist_files = self._get_whitelist_files(whitelist_files, self.blacklist_files)
        self.snippet_marker = snippet_marker
        self.prompt = ""


    def _get_whitelist_files(
        self,
        whitelist_files: List[str],
        blacklist_files: List[str]
    ) -> List[str]:

        matched_files = []

        for root, dirs, files in os.walk(self.cwd):
            relative_root = os.path.relpath(root, self.cwd)
            dirs[:] = [d for d in dirs if not d.startswith('.')
                       and not matches_pattern(os.path.join(relative_root, d), blacklist_files)]

            for file in files:
                file_path = os.path.join(relative_root, file)
                if file.startswith('.'):
                    continue

                if matches_pattern(file_path, blacklist_files):
                    continue
                if matches_pattern(file_path, whitelist_files):
                    matched_files.append(file_path)

        return matched_files

    def _add_project_overview(self, prompt: str, project_overview: str) -> str:
        prompt += f"""
# Project Overview
{project_overview}\n
"""
        return prompt

    def _add_project_structure(self, prompt: str, project_structure: str) -> str:
        prompt += f"""
# Project Structure
{project_structure}\n
"""
        return prompt

    def _get_relative_path(self, file: str) -> str:
        return os.path.relpath(os.path.join(self.cwd, file), self.cwd)

    def _add_file_content(self, prompt: str, file: str, file_content: str) -> str:
        relative_path = self._get_relative_path(file)
        prompt += f"""
{self.snippet_marker}{relative_path}
{file_content}
{self.snippet_marker}\n
"""
        return prompt

    def _add_files_content(self, prompt: str, py_options: Dict) -> str:
        for file in self.whitelist_files:
            file_content = get_file_content(file, py_options)
            prompt = self._add_file_content(prompt, file, file_content)
        return prompt

    def project_info(
            self,
            project_overview: str = "",
            add_project_structure: bool = False,
            add_files_content: bool = False,
            py_options: Union[Dict, None] = None,
    ) -> str:
        """
        Generate the project information for the prompt

        Parameters:
        ----------
        project_overview : str, optional
            The project overview, by default ""
        add_project_structure : bool, optional
            Whether to include the project structure in the prompt, by default False
        add_files_content : bool, optional
            Whether to include the content of the project files in the prompt, by default False
        py_options : Union[Dict, None], optional
            The Python options to customize the content of the Python files, by default None

        Returns:
        -------
        str
            The prompt generated by the `Damj` class
        """
        if py_options is None:
            py_options = {
                "add_comments": True,
                "add_imports": True,
                "add_docstrings": True,
                "ipynb_output": False
            }

        prompt = self.prompt
        if project_overview:
            prompt = self._add_project_overview(self.prompt, project_overview)

        if add_project_structure:
            project_structure_str = get_project_structure(self.cwd, self.blacklist_files)
            prompt = self._add_project_structure(prompt, project_structure_str)

        if add_files_content:
            prompt = self._add_files_content(prompt, py_options)

        self.prompt = prompt
        return prompt


    def _add_question(self, prompt: str, question: str) -> str:
        prompt += f"""
# Question
{question}\n
"""
        return prompt

    def _post_process_prompt(self, prompt: str) -> str:
        prompt = prompt.strip()
        return prompt

    def create_prompt(
        self,
        question: Union[str, None] = None,
        copy_to_clipboard: bool = True,
        to_markdown: bool = False,
    ) -> Union[str, Markdown]:
        """
        Create the prompt for the large language model (LLM)

        Parameters:
        ----------
        question : Union[str, None], optional
            The question to include in the prompt, by default None
        copy_to_clipboard : bool, optional
            Whether to copy the prompt to the clipboard, by default True
        to_markdown : bool, optional
            Whether to display the prompt as Markdown, by default False

        Returns:
        -------
        Union[str, Markdown]
            The prompt generated by the `Damj` class
        """
        prompt = self.prompt
        if question:
            prompt = self._add_question(prompt, question)

        prompt = self._post_process_prompt(prompt)

        if copy_to_clipboard:
            pyperclip.copy(prompt)

        if to_markdown:
            return show_markdown(prompt)

        return prompt



"""
Utility functions for the DAMJ package

This module provides utility functions for the DAMJ package, including functions to get
the project structure, get the content of a file, and convert text to markdown.
"""
import os
import ast
import json
import textwrap
from typing import List
from IPython.display import Markdown


def get_indent(level: int) -> str:
    """
    Get the indentation for a given level

    Parameters:
    ----------
    level : int
        The level of indentation

    Returns:
    -------
    str
        The indentation string
    """
    return "|   " * level

def matches_pattern(file_path: str, patterns: List[str]) -> bool:
    """
    Check if a file path matches any of the patterns

    Parameters:
    ----------
    file_path : str
        The file path to check
    patterns : List[str]
        The list of patterns to match

    Returns:
    -------
    bool
        True if the file path matches any of the patterns, False otherwise
    """
    for pattern in patterns:
        if pattern == "*":
            return True
        if pattern in file_path:
            return True
    return False

def get_project_structure(cwd: str, blacklist_files: List[str]=None) -> str:
    """
    Get the project structure

    Parameters:
    ----------
    cwd : str
        The current working directory
    blacklist_files : List[str]
        The list of files to exclude

    Returns:
    -------
    str
        The project structure
    """
    project_structure_str = ""
    if blacklist_files is None:
        blacklist_files = []
    for root, dirs, files in os.walk(cwd):
        dirs[:] = [d for d in dirs
                   if not d.startswith('.')
                   and not matches_pattern(os.path.join(root, d), blacklist_files)]

        current_dir = os.path.relpath(root, cwd)
        indent_level = current_dir.count(os.sep)
        indent = get_indent(indent_level)

        if current_dir != ".":
            project_structure_str += f"{indent}├── {os.path.basename(root)}/\n"

        files.sort()

        for file in files:
            if file.startswith('.'):
                continue
            if matches_pattern(os.path.join(current_dir, file), blacklist_files):
                continue
            file_indent = get_indent(indent_level + 1)
            project_structure_str += f"{file_indent}├── {file}\n"

    return project_structure_str

def handle_ipynb(file_path: str, py_options: dict) -> str:
    """
    Handle Jupyter notebooks content

    Parameters:
    ----------
    file_path : str
        The file path of the Jupyter notebook
    py_options : dict
        The Python options
        The options include:
            - add_comments: bool
                Whether to include comments in the code
            - add_imports: bool
                Whether to include imports in the code
            - add_docstrings: bool
                Whether to include docstrings in the code
            - ipynb_output: bool
                Whether to include the output of the Jupyter notebook

    Returns:
    -------
    str
        The processed code
    """
    with open(file_path, "r", encoding="utf-8") as file:
        notebook_content = json.load(file)

    result = ""
    add_comments = py_options.get("add_comments", True)
    add_imports = py_options.get("add_imports", True)
    add_docstrings = py_options.get("add_docstrings", True)
    include_output = py_options.get("ipynb_output", False)

    def process_code(code: str) -> str:
        """
        Process the code

        Parameters:
        ----------
        code : str
            The code to process

        Returns:
        -------
        str
            The processed code
        """
        if not add_docstrings:
            tree = ast.parse(code)
            tree = strip_docstrings(tree)
            code = ast.unparse(tree)

        if not add_comments:
            code = "\n".join(line for line in code.splitlines()
                             if not line.strip().startswith("#"))

        if not add_imports:
            code = "\n".join(
                line for line in code.splitlines()
                if not line.strip().startswith("import")
                and not line.strip().startswith("from")
            )

        return code

    for cell in notebook_content.get("cells", []):
        if cell.get("cell_type") == "code":
            cell_code = "".join(cell.get("source", []))
            processed_code = process_code(cell_code)
            result += processed_code + "\n"
            if include_output:
                for output in cell.get("outputs", []):
                    if "text" in output:
                        result += "".join(output["text"]) + "\n"
                    elif "data" in output and "text/plain" in output["data"]:
                        result += "".join(output["data"]["text/plain"]) + "\n"
    return result

def strip_docstrings(node: ast.AST) -> ast.AST:
    """
    Strip docstrings from the AST

    Parameters:
    ----------
    node : ast.AST
        The AST node

    Returns:
    -------
    ast.AST
        The AST node with docstrings stripped
    """
    if isinstance(node, (ast.Module, ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)):
        node.body = [n for n in node.body if not (isinstance(n, ast.Expr)
                                            and isinstance(n.value, ast.Str))]
    for child in ast.iter_child_nodes(node):
        strip_docstrings(child)
    return node


def get_file_content(file: str, py_options: dict) -> str:
    """
    Get the content of a file

    Parameters:
    ----------
    file : str
        The file path
    py_options : dict
        The Python options
        The options include:
            - add_comments: bool
                Whether to include comments in the code
            - add_imports: bool
                Whether to include imports in the code
            - add_docstrings: bool
                Whether to include docstrings in the code

    Returns:
    -------
    str
        The content of the file
    """
    # GET OPTIONS
    add_comments = py_options.get("add_comments", True)
    add_imports = py_options.get("add_imports", True)
    add_docstrings = py_options.get("add_docstrings", True)

    # HANDLE JUPYTER NOTEBOOKS
    if file.endswith(".ipynb"):
        return handle_ipynb(file, py_options)

    # READ FILE
    with open(file, "r", encoding="latin-1") as file_code:
        new_code = file_code.read()

    # REMOVE COMMENTS
    if not add_comments:
        new_code = "\n".join(line for line in new_code.splitlines()
                             if not line.strip().startswith("#"))

    # REMOVE IMPORTS
    if not add_imports:
        new_code = "\n".join(line for line in new_code.splitlines()
                             if not line.strip().startswith("import")
                             and not line.strip().startswith("from"))

    # REMOVE DOCSTRINGS
    if not add_docstrings:
        tree = ast.parse(new_code)
        new_code = ast.unparse(strip_docstrings(tree))

    return new_code


def show_markdown(text: str) -> Markdown:
    """
    Convert text to markdown

    Parameters:
    ----------
    text : str
        The text to convert to markdown

    Returns:
    -------
    Markdown
        The markdown text
    """
    text = text.replace('•', '  *')
    text = text.replace('\n', '  \n')
    return Markdown(textwrap.indent(text, '> ', predicate=lambda _: True))



"""
damj

This module provides the main class `Damj` for the `damj` package. The `Damj` class is designed to
help developers create prompts for large language models (LLMs) with ease by combining different
project files within the developer's working directory. The package supports various customization
options for including or excluding certain types of content from Python files and Jupyter
notebooks.

The `__all__` directive is used to specify the public interface of the module, which includes
the `Damj` class.

Classes:
    Damj: A class to generate prompts for LLMs by processing project files and applying
    customization options.

Usage example:
    ```python
    from damj import Damj

    damj = Damj(
        cwd='/path/to/project',
        whitelist_files=['*.py'],
        blacklist_files=['.venv', '__pycache__']
    )

    prompt = damj.project_info(
        project_overview="This is a sample project.",
        add_project_structure=True,
        add_files_content=True,
        py_options={'add_imports': True,
                    'add_comments': True,
                    'add_docstrings': False,
                    'ipynb_output': False}
    )

    prompt = damj.create_prompt(
        question="What is the purpose of this project?",
        copy_to_clipboard=True,
        to_markdown=False,
    )

    print(prompt)
    ```
"""
from damj.damj import Damj
__all__ = ['Damj']
